import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer

from requests_oauthlib import OAuth1Session
from requests_oauthlib.oauth1_session import TokenMissing, TokenRequestDenied

from evernote_backup.cli_app_util import is_inside_docker
from evernote_backup.evernote_client import EvernoteClientBase


class OAuthDeclinedError(Exception):
    """Raise when user cancels authentication"""


class CallbackHandler(BaseHTTPRequestHandler):
    http_codes = {
        "OK": 200,
        "NOT FOUND": 404,
    }

    def do_GET(self) -> None:
        if not self.path.startswith("/oauth_callback?"):
            self.send_response(self.http_codes["NOT FOUND"])
            self.end_headers()
            return

        self.server.callback_response = self.path

        self.send_response(self.http_codes["OK"])
        self.end_headers()
        self.wfile.write(
            "<html><head><title>OAuth Callback</title></head>"
            "<body>You can close this tab now...</body></html>".encode("utf-8")
        )

    def log_message(self, *args, **kwargs) -> None:  # type: ignore
        """Silencing server log"""


class StoppableHTTPServer(HTTPServer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)

        self.callback_response: str = ""

    def run(self) -> None:
        try:  # noqa: WPS501
            self.serve_forever()
        finally:
            self.server_close()


class EvernoteOAuthCallbackHandler(object):
    def __init__(
        self, oauth_client: "EvernoteOAuthClient", oauth_port: int, server_host: str
    ) -> None:
        self.client = oauth_client

        self.server_host = server_host
        self.server_port = oauth_port

    def get_oauth_url(self) -> str:
        return self.client.get_authorize_url(
            f"http://{self.server_host}:{self.server_port}/oauth_callback"
        )

    def wait_for_token(self) -> str:
        return self.client.get_access_token(self._wait_for_callback())

    def _wait_for_callback(self) -> str:
        if is_inside_docker():
            server_param = ("0.0.0.0", self.server_port)  # noqa: S104
        else:
            server_param = (self.server_host, self.server_port)

        callback_server = StoppableHTTPServer(server_param, CallbackHandler)

        thread = threading.Thread(target=callback_server.run)
        thread.start()

        try:  # noqa: WPS501
            while not callback_server.callback_response:
                time.sleep(0.1)
        finally:
            callback_server.shutdown()
            thread.join()

        return callback_server.callback_response


class EvernoteOAuthClient(EvernoteClientBase):
    def __init__(
        self,
        backend: str,
        consumer_key: str,
        consumer_secret: str,
    ) -> None:
        super().__init__(backend=backend)

        self.client_key = consumer_key
        self.client_secret = consumer_secret

        self._session = None

    def get_authorize_url(self, callback_url: str) -> str:
        self._session = OAuth1Session(
            client_key=self.client_key,
            client_secret=self.client_secret,
            callback_uri=callback_url,
        )

        self._session.fetch_request_token(self._get_endpoint("oauth"))

        return self._session.authorization_url(self._get_endpoint("OAuth.action"))

    def get_access_token(self, callback_response_raw: str) -> str:
        try:
            self._session.parse_authorization_response(callback_response_raw)
        except TokenMissing:
            raise OAuthDeclinedError

        try:
            access_token = self._session.fetch_access_token(self._get_endpoint("oauth"))
        except TokenRequestDenied:
            raise OAuthDeclinedError

        return access_token["oauth_token"]



from pathlib import Path
from typing import Dict, Tuple

MAX_FILE_NAME_LEN = 255


class SafePath(object):
    """Ensures path for tuples of directory names that may contain bad symbols

    Use case:
    safe_path.get("test_dir") == base_dir/test_dir
    safe_path.get("test_dir?|") == base_dir/test_dir__
    safe_path.get("test_dir||") == base_dir/test_dir__ (1)
    """

    def __init__(self, base_dir: Path, overwrite: bool = False) -> None:
        self.safe_paths: Dict[Tuple[str, ...], Path] = {}

        self.main_base_dir = base_dir
        self.overwrite = overwrite

    def get_file(self, *paths: str) -> Path:
        return self._get(*paths, is_dir=False, overwrite=self.overwrite)

    def get(self, *paths: str) -> Path:
        return self._get(*paths, is_dir=True, overwrite=self.overwrite)

    def _get(self, *paths: str, is_dir: bool, overwrite: bool) -> Path:
        if paths in self.safe_paths:
            return self.safe_paths[paths]  # noqa: WPS529

        if len(paths) > 1:
            # Create all parent dirs
            parent_dir = paths[:-1]
            base_dir = self._get(*parent_dir, is_dir=True, overwrite=True)
        else:
            _ensure_path(self.main_base_dir)
            base_dir = self.main_base_dir

        this_path = paths[-1]

        safe_path = _get_safe_path(base_dir, this_path, overwrite)

        if is_dir:
            safe_path.mkdir(exist_ok=True)
            self.safe_paths[paths] = safe_path

        return safe_path


def _get_safe_path(target_dir: Path, new_name: str, overwrite: bool = False) -> Path:
    file_name = _replace_bad_characters(new_name)
    file_name = _trim_name(Path(file_name).stem, Path(file_name).suffix)

    if not overwrite:
        file_name = _get_non_existant_name(file_name, target_dir)

    return target_dir / file_name


def _ensure_path(output_path: Path) -> None:
    if not output_path.is_dir():
        output_path.mkdir(parents=True)


def _replace_bad_characters(string: str) -> str:
    bad = r'<>:"/\|?*'

    result_string = string
    for b in bad:
        result_string = result_string.replace(b, "_")

    return result_string


def _get_non_existant_name(file_name: str, target_dir: Path) -> str:
    i = 0
    orig = Path(file_name)

    while (target_dir / file_name).exists():
        i += 1

        suffix = f" ({i}){orig.suffix}"
        file_name = f"{orig.stem}{suffix}"

        if len(file_name.encode("utf-8")) > MAX_FILE_NAME_LEN:
            max_len = MAX_FILE_NAME_LEN - len(suffix.encode("utf-8"))

            trimmed_name = _trim_name(orig.stem, "", max_len)
            file_name = f"{trimmed_name}{suffix}"

    return file_name


def _trim_name(file_name: str, file_ext: str, max_len: int = MAX_FILE_NAME_LEN) -> str:
    """Trim file name to 255 characters while maintaining extension
    255 characters is max file name length on linux and macOS
    Windows has a filename limit of 260 bytes

    Assuming character is one byte
    Trimming only stem, extensions stays untouched

    max_len: if provided, trims to this length otherwise MAX_FILE_NAME_LEN

    Raises: ValueError if the file name is too long and cannot be trimmed
    """
    orig_name = f"{file_name}{file_ext}"

    if len(orig_name.encode("utf-8")) <= max_len:
        return orig_name

    if file_ext:
        max_len_name = max_len - len(file_ext.encode("utf-8"))
    else:
        max_len_name = max_len

    trimmed_name = _trim_string(file_name, max_len_name)

    result_name = f"{trimmed_name}{file_ext}"

    if len(result_name.encode("utf-8")) > max_len:
        raise ValueError(
            f"File name is too long but cannot be safely trimmed: {file_name}"
        )

    return result_name


def _trim_string(string: str, max_len: int) -> str:
    if max_len <= 0:
        return ""

    chars = list(string)

    while sum(len(c.encode("utf-8")) for c in chars) > max_len:
        chars.pop(-1)

    return "".join(chars)



import base64
import io
import os
import sys
from typing import Iterable, Iterator, List, Optional, Sequence, TextIO

import click


class ProgramTerminatedError(Exception):
    """Terminate program with an error"""


def unscramble(scrambled_data: bytes) -> List[str]:
    scrambled_data = base64.b64decode(scrambled_data)

    unscrambled = b""
    for i, char in enumerate(scrambled_data):
        xor = len(scrambled_data) - i
        unscrambled += (char ^ xor).to_bytes(1, byteorder="big")

    return unscrambled.decode().split()


def get_progress_output() -> Optional[TextIO]:
    is_verbose_mode = click.get_current_context().find_root().params["verbose"]

    if not is_console_interactive() or is_verbose_mode:
        return io.StringIO()

    return None


def is_console_interactive() -> bool:
    is_quiet = click.get_current_context().find_root().params["quiet"]

    return is_output_to_terminal() and not is_quiet


def is_output_to_terminal() -> bool:
    return sys.stdout.isatty()


def is_inside_docker() -> bool:
    return os.environ.get("INSIDE_DOCKER_CONTAINER", False) is not False


def chunks(lst: Sequence, n: int) -> Iterator[Iterable]:
    """Yield successive n-sized chunks from lst."""

    yield from (lst[i : i + n] for i in range(0, len(lst), n))  # noqa: WPS221
